'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.AttributeBuilder = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

exports.getAttValue = getAttValue;
exports.setAttValue = setAttValue;
exports.attribute = attribute;

var _types = require('../helper/types.js');

var _objects = require('../helper/objects.js');

var _converters = require('../helper/converters.js');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Get the value from an attribute.
 * @param {!HTMLElement} el an HTML element
 * @param {!string} attrName the name of the attribute
 * @param {!boolean} isBoolean true is the returned value should be a boolean
 * @returns {string|boolean}
 */
function getAttValue(el, attrName, isBoolean) {
    if (isBoolean) {
        return el.hasAttribute(attrName);
    }
    return el.getAttribute(attrName);
}

/**
 * Set the value of an attribute.
 * @param {!HTMLElement} el an HTML element
 * @param {!string} attrName the name of the attribute
 * @param {!boolean} isBoolean true is the value should be a boolean
 * @param {string|boolean} value the value to set
 */
function setAttValue(el, attrName, isBoolean, value) {
    if (isBoolean) {
        // Handle boolean value
        if (value && !el.hasAttribute(attrName)) {
            el.setAttribute(attrName, '');
        } else if (!value && el.hasAttribute(attrName)) {
            el.removeAttribute(attrName);
        }
    } else {
        // Handle none boolean value
        if (((0, _types.isUndefined)(value) || (0, _types.isNull)(value)) && el.hasAttribute(attrName)) {
            // There is no value, so the attribute must be removed
            el.removeAttribute(attrName);
        } else if (!(0, _types.isUndefined)(value) && !(0, _types.isNull)(value) && el.getAttribute(attrName) !== value) {
            // Sync the attribute value with value
            el.setAttribute(attrName, value);
        }
    }
}

function getterFactory(attrName, isBoolean) {
    return function () {
        return getAttValue(this, attrName, isBoolean);
    };
}

function setterFactory(attrName, isBoolean, attSetter) {
    return function (value) {
        var attValue = (0, _types.isFunction)(attSetter) ? attSetter.call(this, this, value) : value;
        return setAttValue(this, attrName, isBoolean, attValue);
    };
}

var DEFAULT_DATA = {
    bound: true,
    getterFactory: getterFactory,
    setterFactory: setterFactory,
    getAttValue: getAttValue,
    setAttValue: setAttValue
};

/**
 * The attribute builder.
 * Its goal is to provide a way to define an attribute.
 */

var AttributeBuilder = function () {

    /**
     * @param {!string} attrName the name of the attribute
     */

    function AttributeBuilder(attrName) {
        _classCallCheck(this, AttributeBuilder);

        /**
         * @ignore
         */
        this.data = (0, _objects.assign)({
            attrName: attrName,
            propName: (0, _converters.toCamelCase)(attrName),
            listeners: []
        }, DEFAULT_DATA);
    }

    /**
     * To handle the attribute/property value as a boolean:
     * Attribute is present when true and missing when false.
     * @returns {AttributeBuilder} the builder
     */


    _createClass(AttributeBuilder, [{
        key: 'boolean',
        value: function boolean() {
            this.data.boolean = true;
            return this;
        }

        /**
         * To hide the property name when using <code>Object.keys()</code>.
         * @returns {PropertyBuilder} the builder
         */

    }, {
        key: 'hidden',
        value: function hidden() {
            this.data.enumerable = false;
            return this;
        }

        /**
         * To skip the link between the attribute and its property
         * @returns {AttributeBuilder} the builder
         */

    }, {
        key: 'unbound',
        value: function unbound() {
            this.data.bound = false;
            return this;
        }

        /**
         * To override the property name.
         * @param {!string} propName the property name
         * @returns {AttributeBuilder} the builder
         */

    }, {
        key: 'property',
        value: function property(propName) {
            this.data.propName = propName;
            return this;
        }

        /**
         * To set a default value.
         * @param {*} value the default value
         * @returns {PropertyBuilder} the builder
         */

    }, {
        key: 'value',
        value: function value(_value) {
            this.data.value = _value;
            return this;
        }

        /**
         * To be notified when the attribute is updated.
         * @param {function(el: HTMLElement, oldVal: string, newVal: string)} listener the listener function
         * @returns {AttributeBuilder} the builder
         */

    }, {
        key: 'listen',
        value: function listen(listener) {
            this.data.listeners.push(listener);
            return this;
        }

        /**
         * Logic of the builder.
         * @param {!ElementBuilder.context.proto} proto the prototype
         * @param {!ElementBuilder.on} on the method on
         */

    }, {
        key: 'build',
        value: function build(proto, on) {
            var data = this.data,
                defaultValue = (0, _objects.result)(this.data, 'value'),
                descriptor = {
                enumerable: this.data.enumerable,
                configurable: false,
                get: this.data.getterFactory(this.data.attrName, this.data.boolean),
                set: this.data.setterFactory(this.data.attrName, this.data.boolean)
            };

            if (data.bound) {
                Object.defineProperty(proto, data.propName, descriptor);
            }

            on('after:createdCallback').invoke(function (el) {
                if (data.bound) {
                    var attrValue = getAttValue(el, data.attrName, data.boolean);
                    if (data.boolean) {
                        el[data.propName] = !!defaultValue ? defaultValue : attrValue;
                    } else if (!(0, _types.isNull)(attrValue) && !(0, _types.isUndefined)(attrValue)) {
                        el[data.propName] = attrValue;
                    } else if (!(0, _types.isUndefined)(defaultValue)) {
                        el[data.propName] = defaultValue;
                    }
                }
                if (data.listeners.length > 0) {
                    (function () {
                        var oldValue = data.boolean ? false : null;
                        var setValue = el[data.propName];
                        if (oldValue !== setValue) {
                            data.listeners.forEach(function (listener) {
                                return listener.call(el, el, oldValue, setValue);
                            });
                        }
                    })();
                }
            });

            on('before:attributeChangedCallback').invoke(function (el, attName, oldVal, newVal) {
                // Synchronize the attribute value with its properties
                if (attName === data.attrName) {
                    if (data.bound) {
                        var newValue = data.boolean ? newVal === '' : newVal;
                        if (el[data.propName] !== newValue) {
                            el[data.propName] = newValue;
                        }
                    }
                    if (data.listeners.length > 0) {
                        (function () {
                            var oldValue = data.boolean ? oldVal === '' : oldVal;
                            var setValue = data.boolean ? newVal === '' : newVal;
                            if (oldValue !== setValue) {
                                data.listeners.forEach(function (listener) {
                                    return listener.call(el, el, oldValue, setValue);
                                });
                            }
                        })();
                    }
                }
            });
        }
    }]);

    return AttributeBuilder;
}();

/**
 * Get a new attribute builder.
 * @param {!string} attrName the name of the attribute
 * @returns {AttributeBuilder} the attribute builder
 */


exports.AttributeBuilder = AttributeBuilder;
function attribute(attrName) {
    return new AttributeBuilder(attrName);
}